import os
import pickle
import pytest
import numpy as np
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.datasets import fetch_20newsgroups
from sklearn import metrics
from scipy.sparse import csr_matrix
from sib import SIB


base_dir = os.path.dirname(os.path.abspath(__file__))

n_cls = 4

setups = {
    'baseline':      {'n_cls': n_cls, 'n_jobs': 1, 'n_init': 1, 'max_iter': 8,
                      'tol': 0.05, 'uniform': True, 'sparse': True, 'opt': 'B'},
    'n_init':        {'n_cls': n_cls, 'n_jobs': 1, 'n_init': 8, 'max_iter': 8,
                      'tol': 0.05, 'uniform': True, 'sparse': True, 'opt': 'C'},
    'n_jobs':        {'n_cls': n_cls, 'n_jobs': -1, 'n_init': 8, 'max_iter': 8,
                      'tol': 0.05, 'uniform': True, 'sparse': True, 'opt': 'C'},
    'max_iter':      {'n_cls': n_cls, 'n_jobs': 1, 'n_init': 1, 'max_iter': 3,
                      'tol': 0.02, 'uniform': True, 'sparse': True, 'opt': 'C'},
    'tol':           {'n_cls': n_cls, 'n_jobs': 1, 'n_init': 1, 'max_iter': 8,
                      'tol': 0.10, 'uniform': True, 'sparse': True, 'opt': 'B'},
    'uniform_prior': {'n_cls': n_cls, 'n_jobs': 1, 'n_init': 1, 'max_iter': 8,
                      'tol': 0.05, 'uniform': False, 'sparse': True, 'opt': 'C'},
    'float':         {'n_cls': n_cls, 'n_jobs': 1, 'n_init': 1, 'max_iter': 8,
                      'tol': 0.05, 'uniform': False, 'sparse': True, 'opt': 'C', 'type': 'float'},
    'dense':          {'n_cls': n_cls, 'n_jobs': 1, 'n_init': 1, 'max_iter': 4,
                       'tol': 0.05, 'uniform': True, 'sparse': False, 'opt': 'B'},
}

equal_refs = [['baseline', 'dense'], ['n_init', 'n_jobs'], ['uniform_prior', 'float']]

vectors_path = os.path.join(base_dir, 'resources', 'vectors')

gold_labels_path = os.path.join(base_dir, 'resources', 'gold_labels')

ami_scores_path = os.path.join(base_dir, 'resources', 'ami_scores')

random_state = 527802

max_features = 2500


def get_names():
    return setups.keys()


def create(setup_name, optimizer=None):
    setup = setups[setup_name]
    n_clusters = setup['n_cls']
    n_jobs = setup['n_jobs']
    n_init = setup['n_init']
    max_iter = setup['max_iter']
    tol = setup['tol']
    uniform_prior = setup['uniform']
    optimizer = setup['opt'] if optimizer is None else optimizer
    return SIB(n_clusters=n_clusters, n_jobs=n_jobs, n_init=n_init,
               max_iter=max_iter, tol=tol, uniform_prior=uniform_prior,
               random_state=random_state, optimizer_type=optimizer, verbose=True)


def is_sparse(setup_name):
    setup = setups[setup_name]
    return setup['sparse']


def is_float(setup_name):
    setup = setups[setup_name]
    return 'type' in setup and setup['type'] == 'float'


def get_path(setup_name):
    return os.path.join(base_dir, 'resources', setup_name + '_ref')


def exists(setup_name):
    return os.path.exists(get_path(setup_name))


def save(setup_name, labels, costs, score):
    setup_path = get_path(setup_name)
    os.makedirs(setup_path)
    np.savez_compressed(os.path.join(setup_path, "data.npz"),
                        labels=labels, costs=costs, score=score)


def load(setup_name):
    setup_path = get_path(setup_name)
    data = np.load(os.path.join(setup_path, "data.npz"))
    return data['labels'], data['costs'], data['score']


def generate(setup_name, vectors):
    if not exists(setup_name):
        print("Generating reference for: %s" % setup_name)
        if not is_sparse(setup_name):
            vectors = vectors.toarray()
        if is_float(setup_name):
            vectors = vectors.astype(np.float)
        sib = create(setup_name, optimizer='B')
        sib.fit(vectors)
        save(setup_name, sib.labels_, sib.costs_, sib.score_)
    else:
        print("Reference already exists for " + setup_name)


def verify(setup_name, vectors):
    if exists(setup_name):
        ref_labels, ref_costs, ref_score = load(setup_name)
        if not is_sparse(setup_name):
            vectors = vectors.toarray()
        sib = create(setup_name)
        sib.fit(vectors)
        assert np.allclose(ref_labels, sib.labels_)
        assert np.allclose(ref_costs, sib.costs_)
        assert np.allclose(ref_score, sib.score_)


def vectorize_20ng():
    categories = ['soc.religion.christian', 'comp.graphics',
                  'rec.autos', 'sci.med']
    dataset = fetch_20newsgroups(subset='train', categories=categories)
    vectorizer = CountVectorizer(max_features=max_features)
    return vectorizer.fit_transform(dataset.data), dataset.target


def load_vectors():
    data = np.load(os.path.join(vectors_path, "data.npz"))
    return csr_matrix((data['data'], data['indices'], data['indptr']))


@pytest.fixture
def supply_vectors():
    return load_vectors()


def generate_references():
    vectors, gold_labels = vectorize_20ng()
    for setup_name in setups.keys():
        generate(setup_name, vectors)
    if not os.path.exists(vectors_path):
        os.makedirs(vectors_path)
        print("Saving vectors to: %s" % vectors_path)
        np.savez_compressed(os.path.join(vectors_path, "data.npz"),
                            data=vectors.data, indptr=vectors.indptr, indices=vectors.indices)
    else:
        print("Vectors dump is already generated")

    if not os.path.exists(gold_labels_path):
        os.makedirs(gold_labels_path)
        print("Saving gold labels to: %s" % gold_labels_path)
        np.savez_compressed(os.path.join(gold_labels_path, "gold_labels.npz"), labels=gold_labels)
    else:
        print("Gold labels are already saved")

    # verification that setups that are expected to give equal result really do so
    for setup_list in equal_refs:
        if len(setup_list) > 0:
            first_labels, first_costs, first_score = load(setup_list[0])
            for setup_id, setup_name in enumerate(setup_list):
                if setup_id > 0:
                    current_labels, current_costs, current_score = load(setup_name)
                    assert np.allclose(first_labels, current_labels)
                    assert np.allclose(first_costs, current_costs)
                    assert np.allclose(first_score, current_score)
    print("Verified equal result for marked tests")

    if not os.path.exists(ami_scores_path):
        os.makedirs(ami_scores_path)
        ami_scores = {}
        for setup in setups.keys():
            labels, _, _ = load(setup)
            ami_scores[setup] = metrics.adjusted_mutual_info_score(gold_labels, labels)
        print("Saving ami scores to: %s" % ami_scores_path)
        with open(os.path.join(ami_scores_path, "ami_scores.npz"), "wb") as f:
            pickle.dump(ami_scores, f)
    else:
        with open(os.path.join(ami_scores_path, "ami_scores.npz"), "rb") as f:
            ami_scores = pickle.load(f)
        for setup, ref_ami in ami_scores.items():
            labels, _, _ = load(setup)
            ami = metrics.adjusted_mutual_info_score(gold_labels, labels)
            assert np.isclose(ami, ref_ami)
        print("Verified ami scores")


@pytest.mark.parametrize("setup_name", list(setups.keys()))
def test_setups(supply_vectors, setup_name):
    vectors = supply_vectors
    verify(setup_name, vectors)



# © Copyright IBM Corporation 2020.
#
# LICENSE: Apache License 2.0 (Apache-2.0)
# http://www.apache.org/licenses/LICENSE-2.0

import numpy as np
from scipy.sparse import issparse
from scipy.special import xlogy


class PSIBOptimizer:

    def __init__(self, n_clusters, n_features, n_samples, xy, xy_sum, x_sum):
        self.n_clusters = n_clusters
        self.n_features = n_features
        self.n_samples = n_samples
        self.xy = xy
        self.xy_sum = xy_sum
        self.x_sum = x_sum
        self.sparse = issparse(xy)

    def init_centroids(self, labels, x_ignore, t_size, t_sum, t_log_sum, t_centroid,
                       t_centroid_log_t_centroid, t_centroid_log_t_centroid_sum):
        for i in range(self.n_samples):
            if x_ignore[i]:
                continue
            t = labels[i]
            t_size[t] += 1
            t_sum[t] += self.x_sum[i]
            if self.sparse:
                i_start = self.xy.indptr[i]
                i_end = self.xy.indptr[i + 1]
                v_indices = self.xy.indices[i_start:i_end]
                v_data = self.xy.data[i_start:i_end]
                t_centroid[t, v_indices] += v_data
            else:
                t_centroid[t, :] += self.xy[i, :]
        np.log2(t_sum, out=t_log_sum)
        t_centroid_log_t_centroid[:] = xlogy(t_centroid, t_centroid) / np.log(2)
        if not self.sparse:
            t_centroid_log_t_centroid_sum[:] = np.sum(t_centroid_log_t_centroid, axis=1)

    def optimize(self, x_permutation, t_size, t_sum, t_log_sum, t_centroid,
                 t_centroid_log_t_centroid, t_centroid_log_t_centroid_sum,
                 labels, x_locked_in, ity, ref_labels=None):
        return self.iterate(True, self.n_samples, self.xy, self.xy_sum, self.x_sum,
                            x_permutation, t_size, t_sum, t_log_sum, t_centroid,
                            t_centroid_log_t_centroid, t_centroid_log_t_centroid_sum,
                            labels, x_locked_in, None, ity, ref_labels)

    def infer(self, n_samples, xy, xy_sum, x_sum, t_size, t_sum, t_log_sum, t_centroid,
              t_centroid_log_t_centroid, t_centroid_log_t_centroid_sum,
              labels, x_locked_in, costs, ref_labels=None):
        return self.iterate(False, n_samples, xy, xy_sum, x_sum, None, t_size,
                            t_sum, t_log_sum, t_centroid, t_centroid_log_t_centroid,
                            t_centroid_log_t_centroid_sum, labels, x_locked_in,
                            costs, None, ref_labels)

    def iterate(self, clustering_mode, n_samples, xy, xy_sum, x_sum,
                x_permutation, t_size, t_sum, log_t_sum, t_centroid,
                t_centroid_log_t_centroid, t_centroid_log_t_centroid_sum,
                labels, x_locked_in, costs, ity, ref_labels=None):

        n_changes = 0

        total_cost = 0

        if self.sparse:
            # log_t_sum_plus_x_sum = None
            log_t_centroid_plus_x = None
            # log_t_centroid = None
        else:
            # log_t_sum_plus_x_sum = np.empty_like(t_sum, dtype=float)
            log_t_centroid_plus_x = np.empty_like(t_centroid, dtype=float)
            # log_t_centroid = np.zeros_like(t_centroid, dtype=float)
            # np.log2(t_centroid, where=t_centroid > 0, out=log_t_centroid)

        for i in range(n_samples):
            x = x_permutation[i] if x_permutation is not None else i

            # if this element is already locked-in, we skip it
            if x_locked_in[x]:
                continue

            old_t = labels[x]

            # if t is a singleton cluster we do not reduce it any further
            if clustering_mode and t_size[old_t] == 1:
                continue

            # obtain local references
            if self.sparse:
                x_start = xy.indptr[x]
                x_end = xy.indptr[x + 1]
                x_indices = xy.indices[x_start:x_end]
                x_data = xy.data[x_start:x_end]
            else:
                x_indices = None
                x_data = xy[x, :]
            x_sum_x = x_sum[x]

            if clustering_mode:
                # draw x out of its current cluster
                t_size[old_t] -= 1
                t_sum[old_t] -= x_sum_x
                log_t_sum[old_t] = np.log2(t_sum[old_t])
                if self.sparse:
                    t_centroid[old_t, x_indices] -= x_data
                    t_centroid_log_t_centroid[old_t, x_indices] = xlogy(np.clip(t_centroid[old_t, x_indices],
                                                                                a_min=0, a_max=None),
                                                                        t_centroid[old_t, x_indices]) / np.log(2)
                else:
                    t_centroid[old_t, :] -= x_data
                    t_centroid_log_t_centroid_sum[old_t] = np.sum(xlogy(np.clip(t_centroid[old_t, :],
                                                                                a_min=0, a_max=None),
                                                                        t_centroid[old_t, :])) / np.log(2)

            if self.sparse:
                t_centroid_log_t_centroid_sum = np.sum(t_centroid_log_t_centroid[:, x_indices], axis=1)

            # argmin Df
            t_sum_plus_x_sum = t_sum + x_sum_x

            if self.sparse:
                t_centroid_plus_x = t_centroid[:, x_indices] + x_data
                h_m_plus_t = (t_centroid_plus_x[:, None, :] @
                              np.log2(t_centroid_plus_x)[..., None]).ravel()

            else:
                t_centroid_plus_x = t_centroid + x_data
                h_m_plus_t = np.sum(xlogy(np.clip(t_centroid_plus_x, a_min=0, a_max=None),
                                          t_centroid_plus_x), axis=1) / np.log(2)

            h_t = t_centroid_log_t_centroid_sum

            tmp_costs = -h_m_plus_t + h_t - t_sum * log_t_sum + t_sum_plus_x_sum * np.log2(t_sum_plus_x_sum)
            tmp_costs /= xy_sum

            new_t = np.argmin(tmp_costs).item()
            if ref_labels is not None:
                ref_t = ref_labels[x]
                if new_t != ref_t and not np.isclose(tmp_costs[new_t], tmp_costs[ref_t]):
                    print("override t of cost=%.8f, with cost=%.8f" % (tmp_costs[new_t], tmp_costs[ref_t]))
                    new_t = ref_t

            if clustering_mode:
                ity += tmp_costs[old_t] - tmp_costs[new_t]

                # add x to its new cluster
                t_size[new_t] += 1
                t_sum[new_t] += x_sum_x
                log_t_sum[new_t] = np.log2(t_sum[new_t])

                if self.sparse:
                    t_centroid[new_t, x_indices] += x_data

                    t_centroid_new_t_x = t_centroid[new_t, x_indices]
                    t_centroid_log_t_centroid[new_t, x_indices] = xlogy(t_centroid_new_t_x,
                                                                        t_centroid_new_t_x) / np.log(2)
                    # no need to update t_centroid_log_t_centroid_sum because in the sparse case it is computed
                    # per x, and will be re-computed for the next x when the for-loop iterates
                else:
                    t_centroid_new_t = t_centroid[new_t, :]
                    t_centroid_new_t += x_data

                    t_centroid_log_t_centroid_sum[new_t] = np.sum(xlogy(np.clip(t_centroid_new_t, a_min=0, a_max=None),
                                                                        t_centroid_new_t)) / np.log(2)

                if new_t != old_t:
                    # update the changes counter
                    n_changes += 1
            else:
                total_cost += tmp_costs[new_t]
                costs[x, :] = tmp_costs

            labels[x] = new_t

        if clustering_mode:
            log_sum_xy = np.log2(xy_sum)
            ht = -np.dot(t_sum, log_t_sum - log_sum_xy) / xy_sum
            return n_changes / self.n_samples if self.n_samples > 0 else 0, ity, ht
        else:
            return total_cost
